#include "netfs.h"
#include <linux/smp_lock.h>

/* From inode.c */
extern ino_t netfs_new_ino(void);

static int netfs_readdir(struct file *file, void *dirent, filldir_t fill) {
	struct netfs_dirent *cur;

	/* Seek to starting position in dir */
	cur = (struct netfs_dirent *) file->f_pos;
	if(!cur) cur = (struct netfs_dirent *) file->f_dentry->d_subdirs;
	if(!cur) return -ENOTDIR;

	lock_kernel();

	/* Return directory entries */
	list_for_each_entry_continue(cur, file->f_dentry->d_subdirs, list) {
		if(fill(dirent, cur->dentry.d_name.name,
			cur->dentry.d_name.len, cur,
			cur->dentry.d_inode->i_ino, DT_UNKNOWN) < 0) break;
		file->f_pos = cur;
	}
	
	unlock_kernel();
	return 0;
}

struct file_operations netfs_dir_fops = {
	.read		= generic_read_dir,
	.readdir	= netfs_readdir
};

/* Create a directory entry */
static struct netfs_dirent *netfs_mkdirent(struct dentry *dir,
	const char *name, struct inode *inode) {

	struct netfs_dirent *ent = kmalloc(sizeof(*ent));
	if(!ent) return NULL;

	ent->dentry.d_name.name = name;
	ent->dentry.d_name.len = strlen(name);
	d_add(&ent->dentry, inode);
	inode->i_nlink++;

	return ent;
}

/** Create a directory */
static int netfs_mkdir(struct inode *parent, struct dentry *dentry, int mode) {
	struct netfs_dirent *ent;
	struct inode *inode = iget(parent->i_sb, netfs_new_ino());

	if(!netfs_mkdirent(dentry, ".", inode)) {
		iput(inode->ino);
		return -ENOSPC;
	}
	if(!netfs_mkdirent(dentry, "..", parent)) {
		iput(inode->ino);
		/* Have to free space taken by "." here !!!!
		 * I am too tired now to find out how */
		return -ENOSPC;
	}

	d_add(dentry, inode);
	
	return 0;
}

/** This is called when the VFS looks up a file in a directory.
 * We simply create any file that is looked up. */
struct dentry *netfs_lookup(struct inode *dir, struct dentry *dentry,
		struct nameidata *nd) {
	
	lock_kernel();

	/* Check name length */
	if(dentry->d_name.len > NAME_MAX) {
		unlock_kernel();
		return ERR_PTR(-ENAMETOOLONG);
	}

	/* Create directory */
	netfs_mkdir(dir, dentry, 0755);

	unlock_kernel();
	return NULL;
}

struct inode_operations netfs_dir_iops = {
	.lookup		= netfs_lookup
};
